@@ -1,9 +1,10 @@
 /*
- *  Copyright 1999-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -18,32 +19,28 @@
  * Description: General purpose map object                                 *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Mladen Turk <mturk@apache.org>                             *
- * Version:     $Revision: 411538 $                                          *
+ * Version:     $Revision: 521770 $                                          *
  ***************************************************************************/
 #ifdef AS400
 #include "apr_xlate.h"
 #endif
 
 #include "jk_global.h"
-#include "jk_map.h"
 #include "jk_pool.h"
-#include "jk_map.h"
 #include "jk_util.h"
+#include "jk_map.h"
 
-#define CAPACITY_INC_SIZE (50)
-#define LENGTH_OF_LINE    (8192)
-
-#ifdef AS400
-#define CASE_MASK 0xbfbfbfbf
-#else
-#define CASE_MASK 0xdfdfdfdf
-#endif
+#define CAPACITY_INC_SIZE   (50)
+#define LENGTH_OF_LINE      (8192)
+#define JK_MAP_RECURSION    (20)
+#define JK_MAP_REFERENCE    (".reference")
+#define JK_MAP_REFERENCE_SZ (strlen(JK_MAP_REFERENCE))
 
 /* Compute the "checksum" for a key, consisting of the first
- * 4 bytes, normalized for case-insensitivity and packed into
- * an int...this checksum allows us to do a single integer
+ * 4 bytes, packed into an int.
+ * This checksum allows us to do a single integer
  * comparison as a fast check to determine whether we can
- * skip a strcasecmp
+ * skip a strcmp
  */
 #define COMPUTE_KEY_CHECKSUM(key, checksum)    \
 {                                              \
@@ -65,7 +62,6 @@
         c = (unsigned int)*++k;                \
         checksum |= c;                         \
     }                                          \
-    checksum &= CASE_MASK;                     \
 }
 
 struct jk_map
@@ -145,7 +141,7 @@ void *jk_map_get(jk_map_t *m, const char *name, const void *def)
         unsigned int key;
         COMPUTE_KEY_CHECKSUM(name, key)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcasecmp(m->names[i], name) == 0) {
+            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
                 rc = m->values[i];
                 break;
             }
@@ -163,7 +159,7 @@ int jk_map_get_id(jk_map_t *m, const char *name)
         unsigned int key;
         COMPUTE_KEY_CHECKSUM(name, key)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcasecmp(m->names[i], name) == 0) {
+            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
                 rc = i;
                 break;
             }
@@ -182,7 +178,7 @@ const char *jk_map_get_string(jk_map_t *m, const char *name, const char *def)
         unsigned int key;
         COMPUTE_KEY_CHECKSUM(name, key)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcasecmp(m->names[i], name) == 0) {
+            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
                 rc = m->values[i];
                 break;
             }
@@ -238,26 +234,17 @@ double jk_map_get_double(jk_map_t *m, const char *name, double def)
 int jk_map_get_bool(jk_map_t *m, const char *name, int def)
 {
     char buf[100];
-    size_t len;
     const char *rc;
-    int rv = 0;
 
     sprintf(buf, "%d", def);
     rc = jk_map_get_string(m, name, buf);
 
-    len = strlen(rc);
-    if (len) {
-        if (strcasecmp(rc, "true") == 0 ||
-            *rc == 'Y' || *rc == 'y' || *rc == '1') {
-            rv = 1;
-        }
-    }
-    return rv;
+    return jk_get_bool_code(rc, def);
 }
 
 char **jk_map_get_string_list(jk_map_t *m,
                               const char *name,
-                              unsigned *list_len, const char *def)
+                              unsigned int *list_len, const char *def)
 {
     const char *l = jk_map_get_string(m, name, def);
     char **ar = NULL;
@@ -310,6 +297,76 @@ char **jk_map_get_string_list(jk_map_t *m,
     return ar;
 }
 
+int jk_map_get_int_list(jk_map_t *m,
+                        const char *name,
+                        int *list,
+                        unsigned int list_len,
+                        const char *def)
+{
+    const char *l = jk_map_get_string(m, name, def);
+
+#if defined(AS400) || defined(_REENTRANT)
+    char *lasts;
+#endif
+    
+    if (!list_len)
+        return 0;
+
+    if (l) {
+        unsigned int capacity = list_len;
+        unsigned int index = 0;
+        char *p;
+        char *v = jk_pool_strdup(&m->p, l);
+
+        if (!v) {
+            return 0;
+        }
+
+        /*
+         * GS, in addition to VG's patch, we now need to
+         * strtok also by a "*"
+         */
+#if defined(AS400) || defined(_REENTRANT)
+        for (p = strtok_r(v, " \t,", &lasts);
+             p; p = strtok_r(NULL, " \t,", &lasts))
+#else
+        for (p = strtok(v, " \t,"); p; p = strtok(NULL, " \t,"))
+#endif
+
+        {
+            if (index < capacity) {
+                list[index] = atoi(p);
+                index++;
+            }
+            else
+                break;                
+        }
+        return index;
+    }
+    return 0;
+}
+
+int jk_map_add(jk_map_t *m, const char *name, const void *value)
+{
+    int rc = JK_FALSE;
+
+    if (m && name) {
+        unsigned int key;
+        COMPUTE_KEY_CHECKSUM(name, key)
+        map_realloc(m);
+
+        if (m->size < m->capacity) {
+            m->values[m->size] = value;
+            m->names[m->size] = jk_pool_strdup(&m->p, name);
+            m->keys[m->size] = key;
+            m->size++;
+            rc = JK_TRUE;
+        }
+    }
+
+    return rc;
+}
+
 int jk_map_put(jk_map_t *m, const char *name, const void *value, void **old)
 {
     int rc = JK_FALSE;
@@ -319,7 +376,7 @@ int jk_map_put(jk_map_t *m, const char *name, const void *value, void **old)
         unsigned int key;
         COMPUTE_KEY_CHECKSUM(name, key)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcasecmp(m->names[i], name) == 0) {
+            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
                 break;
             }
         }
@@ -331,29 +388,25 @@ int jk_map_put(jk_map_t *m, const char *name, const void *value, void **old)
             rc = JK_TRUE;
         }
         else {
-            map_realloc(m);
-
-            if (m->size < m->capacity) {
-                m->values[m->size] = value;
-                m->names[m->size] = jk_pool_strdup(&m->p, name);
-                m->keys[m->size] = key;
-                m->size++;
-                rc = JK_TRUE;
-            }
+            rc = jk_map_add(m, name, value);
         }
     }
 
     return rc;
 }
 
-int jk_map_read_property(jk_map_t *m, const char *str)
+int jk_map_read_property(jk_map_t *m, const char *str, int allow_duplicates, jk_logger_t *l)
 {
     int rc = JK_TRUE;
     char buf[LENGTH_OF_LINE + 1];
     char *prp = &buf[0];
 
-    if (strlen(str) > LENGTH_OF_LINE)
+    if (strlen(str) > LENGTH_OF_LINE) {
+        jk_log(l, JK_LOG_WARNING,
+               "Line to long (%d > %d), ignoring entry",
+               strlen(str), LENGTH_OF_LINE);
         return JK_FALSE;
+    }
 
     strcpy(prp, str);
     if (trim(prp)) {
@@ -364,36 +417,58 @@ int jk_map_read_property(jk_map_t *m, const char *str)
             trim(prp);
             trim(v);
             if (strlen(v) && strlen(prp)) {
-                const char *oldv = jk_map_get_string(m, prp, NULL);
-                v = jk_map_replace_properties(v, m);
-                if (oldv && jk_is_unique_property(prp) == JK_FALSE) {
-                    char *tmpv = jk_pool_alloc(&m->p,
-                                       strlen(v) + strlen(oldv) + 3);
-                    if (tmpv) {
-                        char sep = '*';
-                        if (jk_is_path_poperty(prp))
-                            sep = PATH_SEPERATOR;
-                        else if (jk_is_cmd_line_poperty(prp))
-                            sep = ' ';
-                        else if (!strcasecmp(prp, "worker.list"))
-                            sep = ',';
-                        sprintf(tmpv, "%s%c%s", oldv, sep, v);
+                const char *oldv;
+                int off = (int)strlen(prp) - (int)JK_MAP_REFERENCE_SZ;
+                /* check the worker properties */
+                if (off <= 0 || strncmp(&prp[off], JK_MAP_REFERENCE, JK_MAP_REFERENCE_SZ) ) {
+                    if (!jk_is_valid_property(prp)) {
+                        jk_log(l, JK_LOG_ERROR,
+                               "The attribute '%s' is not supported - please check"
+                               " the documentation for the supported attributes.",
+                               prp);
+                        return JK_FALSE;
+                    }
+                    if (jk_is_deprecated_property(prp)) {
+                        jk_log(l, JK_LOG_WARNING,
+                               "The attribute '%s' is deprecated - please check"
+                               " the documentation for the correct replacement.",
+                               prp);
+                    }
+                }
+                oldv = jk_map_get_string(m, prp, NULL);
+                v = jk_map_replace_properties(m, v);
+                if (oldv) {
+                    if (allow_duplicates && jk_is_unique_property(prp) == JK_FALSE) {
+                        char *tmpv = jk_pool_alloc(&m->p,
+                                           strlen(v) + strlen(oldv) + 3);
+                        if (tmpv) {
+                            char sep = '*';
+                            if (jk_is_path_property(prp))
+                                sep = PATH_SEPERATOR;
+                            else if (jk_is_cmd_line_property(prp))
+                                sep = ' ';
+                            else if (jk_is_list_property(prp))
+                                sep = ',';
+                            sprintf(tmpv, "%s%c%s", oldv, sep, v);
+                        }
+                        v = tmpv;
+                    }
+                    else {
+                        jk_log(l, JK_LOG_WARNING,
+                               "Duplicate key '%s' detected - previous value '%s'"
+                               " will be overwritten with '%s'.",
+                               prp, oldv ? oldv : "(null)", v ? v : "(null)");
+                        v = jk_pool_strdup(&m->p, v);
                     }
-                    v = tmpv;
                 }
                 else {
-                    if (jk_is_deprecated_property(prp)) {
-                        /* TODO: Log deprecated directive.
-                         * It would require to pass the jk_log_t
-                         * to jk_map_ functions.
-                         */    
-                    }
                     v = jk_pool_strdup(&m->p, v);
                 }
                 if (v) {
                     jk_map_put(m, prp, v, NULL);
                 }
                 else {
+                    JK_LOG_NULL_PARAMS(l);
                     rc = JK_FALSE;
                 }
             }
@@ -403,7 +478,7 @@ int jk_map_read_property(jk_map_t *m, const char *str)
 }
 
 
-int jk_map_read_properties(jk_map_t *m, const char *f, time_t *modified)
+int jk_map_read_properties(jk_map_t *m, const char *f, time_t *modified, int allow_duplicates, jk_logger_t *l)
 {
     int rc = JK_FALSE;
 
@@ -427,7 +502,7 @@ int jk_map_read_properties(jk_map_t *m, const char *f, time_t *modified)
             while (NULL != (prp = fgets(buf, LENGTH_OF_LINE, fp))) {
                 trim_prp_comment(prp);
                 if (*prp) {
-                    if ((rc = jk_map_read_property(m, prp)) == JK_FALSE)
+                    if ((rc = jk_map_read_property(m, prp, allow_duplicates, l)) == JK_FALSE)
                         break;
                 }
             }
@@ -542,7 +617,7 @@ static int map_realloc(jk_map_t *m)
  *  Replace $(property) in value.
  *
  */
-char *jk_map_replace_properties(const char *value, jk_map_t *m)
+char *jk_map_replace_properties(jk_map_t *m, const char *value)
 {
     char *rc = (char *)value;
     char *env_start = rc;
@@ -603,3 +678,206 @@ char *jk_map_replace_properties(const char *value, jk_map_t *m)
 
     return rc;
 }
+
+/**
+ *  Resolve references
+ *
+ */
+int jk_map_resolve_references(jk_map_t *m, const char *prefix,
+                              int wildcard, int depth, jk_logger_t *l)
+{
+    int rc = JK_FALSE;
+
+    JK_TRACE_ENTER(l);
+
+    if (m && prefix) {
+        if (depth <= JK_MAP_RECURSION) {
+            size_t prelen = strlen(prefix);
+            unsigned int i;
+            rc = JK_TRUE;
+            if (JK_IS_DEBUG_LEVEL(l))
+                jk_log(l, JK_LOG_DEBUG,
+                       "Checking for references with prefix %s with%s wildcard (recursion %d)",
+                       prefix, wildcard? "" : "out", depth);
+            for (i = 0; i < m->size; i++) {
+                if (m->values[i] && !strncmp(m->names[i], prefix, prelen)) {
+                    size_t remain = strlen(m->names[i]) - prelen;
+                    if ((remain == JK_MAP_REFERENCE_SZ ) || (wildcard && remain > JK_MAP_REFERENCE_SZ)) {
+                        remain = strlen(m->names[i]) - JK_MAP_REFERENCE_SZ;
+                        if (!strncmp(m->names[i] + remain, JK_MAP_REFERENCE, JK_MAP_REFERENCE_SZ)) {
+                            char *from = jk_pool_alloc(&m->p,
+                                                       (sizeof(char) *
+                                                       (strlen(m->values[i]) + 2)));
+                            char *to = jk_pool_alloc(&m->p,
+                                                     (sizeof(char) *
+                                                     (remain + 2)));
+                            if (!from || !to) {
+                                jk_log(l, JK_LOG_ERROR,
+                                       "Error in string allocation");
+                                rc = JK_FALSE;
+                                break;
+                            }
+                            strcpy(from, m->values[i]);
+                            *(from+strlen(m->values[i]))   = '.';
+                            *(from+strlen(m->values[i])+1) = '\0';
+                            strncpy(to, m->names[i], remain);
+                            *(to+remain)   = '.';
+                            *(to+remain+1) = '\0';
+
+                            rc = jk_map_resolve_references(m, m->values[i], 0, depth+1, l);
+                            if (rc == JK_FALSE) {
+                                break;
+                            }
+                            if (JK_IS_DEBUG_LEVEL(l))
+                                jk_log(l, JK_LOG_DEBUG,
+                                       "Copying values from %s to %s",
+                                       from, to);
+                            rc = jk_map_inherit_properties(m, from, to, l);
+                            if (rc == JK_FALSE) {
+                                break;
+                            }
+                            m->values[i] = NULL;
+                        }
+                    }
+                }
+            }
+        }
+        else {
+            jk_log(l, JK_LOG_ERROR,
+                   "Recursion limit %d for worker references with prefix '%s' reached",
+                   JK_MAP_RECURSION, prefix);
+        }
+    }
+    else {
+        JK_LOG_NULL_PARAMS(l);
+    }
+    JK_TRACE_EXIT(l);
+    return rc;
+}
+
+/**
+ *  Inherit properties
+ *
+ */
+int jk_map_inherit_properties(jk_map_t *m, const char *from, const char *to, jk_logger_t *l)
+{
+    int rc = JK_FALSE;
+    const char *prp;
+    char *to_prp;
+
+    if (m && from && to) {
+        unsigned int i;
+        for (i = 0; i < m->size; i++) {
+            if (!strncmp(m->names[i], from, strlen(from))) {
+                rc = JK_TRUE;
+                prp = m->names[i] + strlen(from);
+                to_prp = jk_pool_alloc(&m->p,
+                                       (sizeof(char) *
+                                       (strlen(to) +
+                                       strlen(prp) + 1)));
+                if (!to_prp) {
+                    jk_log(l, JK_LOG_ERROR,
+                           "Error in string allocation for attribute '%s.%s'",
+                           to, prp);
+                    rc = JK_FALSE;
+                    break;
+                }
+                strcpy(to_prp, to);
+                strcat(to_prp, prp);
+                if (jk_map_get_id(m, to_prp) < 0 ) {
+                    rc = jk_map_add(m, to_prp, m->values[i]);
+                    if (rc == JK_FALSE) {
+                        jk_log(l, JK_LOG_ERROR,
+                               "Error when adding attribute '%s'",
+                               to_prp);
+                        break;
+                    }
+                }
+            }
+        }
+        if ( rc == JK_FALSE) {
+            jk_log(l, JK_LOG_ERROR,
+                   "Reference '%s' not found",
+                   from);
+        }
+    }
+    else {
+        JK_LOG_NULL_PARAMS(l);
+    }
+    return rc;
+}
+
+int jk_map_load_property(jk_map_t *m, const char *str, jk_logger_t *l)
+{
+    int rc = JK_TRUE;
+    char buf[LENGTH_OF_LINE + 1];
+    char *prp = &buf[0];
+
+    if (strlen(str) > LENGTH_OF_LINE) {
+        jk_log(l, JK_LOG_WARNING,
+               "Line to long (%d > %d), ignoring entry",
+               strlen(str), LENGTH_OF_LINE);
+        return JK_FALSE;
+    }
+
+    strcpy(prp, str);
+    if (trim(prp)) {
+        char *v = strchr(prp, '=');
+        if (v) {
+            *v = '\0';
+            v++;
+            trim(prp);
+            trim(v);
+            if (strlen(v) && strlen(prp)) {
+                v = jk_pool_strdup(&m->p, v);
+                if (v) {
+                    jk_map_put(m, prp, v, NULL);
+                }
+                else {
+                    JK_LOG_NULL_PARAMS(l);
+                    rc = JK_FALSE;
+                }
+            }
+        }
+    }
+    return rc;
+}
+
+
+int jk_map_load_properties(jk_map_t *m, const char *f, time_t *modified, jk_logger_t *l)
+{
+    int rc = JK_FALSE;
+
+    if (m && f) {
+        struct stat statbuf;
+        FILE *fp;
+        if ((rc = stat(f, &statbuf)) == -1)
+            return JK_FALSE;
+#ifdef AS400
+        fp = fopen(f, "r, o_ccsid=0");
+#else
+        fp = fopen(f, "r");
+#endif
+
+        if (fp) {
+            char buf[LENGTH_OF_LINE + 1];
+            char *prp;
+
+            rc = JK_TRUE;
+
+            while (NULL != (prp = fgets(buf, LENGTH_OF_LINE, fp))) {
+                trim_prp_comment(prp);
+                if (*prp) {
+                    if ((rc = jk_map_load_property(m, prp, l)) == JK_FALSE)
+                        break;
+                }
+            }
+            fclose(fp);
+            if (modified)
+                *modified = statbuf.st_mtime;
+        }
+    }
+
+    return rc;
+}
+
