@@ -1,9 +1,10 @@
 /*
- *  Copyright 1999-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -21,7 +22,7 @@
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Dan Milstein <danmil@shore.net>                            *
  * Author:      Henri Gomez <hgomez@apache.org>                            *
- * Version:     $Revision: 420266 $                                          *
+ * Version:     $Revision: 483456 $                                          *
  ***************************************************************************/
 
 #ifndef JK_SERVICE_H
@@ -186,12 +187,12 @@ struct jk_ws_service
     unsigned num_attributes;        /* Number of request attributes     */
 
     /*
-     * The jvm route is in use when the adapter load balance among
-     * several JVMs. It is the ID of a specific JVM in the load balance
-     * group. We are using this variable to implement JVM session
+     * The route is in use when the adapter load balance among
+     * several workers. It is the ID of a specific target in the load balance
+     * group. We are using this variable to implement target session
      * affinity
      */
-    const char *jvm_route;
+    const char *route;
 
     /* Temp solution for auth. For native1 it'll be sent on each request,
        if an option is present. For native2 it'll be sent with the first
@@ -200,11 +201,6 @@ struct jk_ws_service
        and it'll work.
      */
     const char *secret;
-    /*
-     * Callbacks into the web server.  For each, the first argument is
-     * essentially a 'this' pointer.  All return JK_TRUE on success
-     * and JK_FALSE on failure.
-     */
 
     /*
      * Area to get POST data for fail-over recovery in POST
@@ -221,10 +217,25 @@ struct jk_ws_service
      */
     int flush_packets;
 
+    /*
+     * If set call flush after AJP13_SEND_HEADERS.
+     */
+    int flush_header;
+
+    /*
+     * HTTP status sent from container.
+     */
+    int http_response_status;
+
     /* Uri worker map. Added for virtual host support
      */
     jk_uri_worker_map_t *uw_map;
 
+    /*
+     * Callbacks into the web server.  For each, the first argument is
+     * essentially a 'this' pointer.  All return JK_TRUE on success
+     * and JK_FALSE on failure.
+     */
     /*
      * Send the response headers to the browser.
      */
@@ -254,6 +265,19 @@ struct jk_ws_service
      * Flush a chunk of response data back to the browser.
      */
     void (JK_METHOD * flush) (jk_ws_service_t *s);
+
+    /*
+     * If set do not reuse socket after each full response
+     */
+    int disable_reuse;
+
+    /*
+     * Add more data to log facilities.
+     */
+    void (JK_METHOD * add_log_items) (jk_ws_service_t *s,
+                                      const char *const *log_names,
+                                      const char *const *log_values,
+                                      unsigned num_of_items);
 };
 
 /*
