@@ -1,9 +1,10 @@
 /*
- *  Copyright 1999-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -19,7 +20,7 @@
  *                                                                         *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Mladen Turk <mturk@apache.org>                             *
- * Version:     $Revision: 420258 $                                          *
+ * Version:     $Revision: 511845 $                                          *
  ***************************************************************************/
 
 #include "jk_pool.h"
@@ -37,6 +38,57 @@
 #endif
 
 
+static const char *uri_worker_map_source_type[] = {
+    "unknown",
+    SOURCE_TYPE_TEXT_WORKERDEF,
+    SOURCE_TYPE_TEXT_JKMOUNT,
+    SOURCE_TYPE_TEXT_URIMAP,
+    SOURCE_TYPE_TEXT_DISCOVER,
+    NULL
+};
+
+
+/* Return the string representation of the uwr source */
+const char *uri_worker_map_get_source(uri_worker_record_t *uwr, jk_logger_t *l)
+{
+    return uri_worker_map_source_type[uwr->source_type];
+}
+
+/* Return the string representation of the uwr match type */
+char *uri_worker_map_get_match(uri_worker_record_t *uwr, char *buf, jk_logger_t *l)
+{
+    unsigned int match;
+
+    buf[0] = '\0';
+    match = uwr->match_type;
+
+    if (match & MATCH_TYPE_DISABLED)
+        strcat(buf, "Disabled ");
+/* deprecated
+    if (match & MATCH_TYPE_STOPPED)
+        strcat(buf, "Stopped ");
+ */
+    if (match & MATCH_TYPE_NO_MATCH)
+        strcat(buf, "Unmount ");
+    if (match & MATCH_TYPE_EXACT)
+        strcat(buf, "Exact");
+    else if (match & MATCH_TYPE_WILDCHAR_PATH)
+        strcat(buf, "Wildchar");
+/* deprecated
+    else if (match & MATCH_TYPE_CONTEXT)
+        strcat(buf, "Context");
+    else if (match & MATCH_TYPE_CONTEXT_PATH)
+        strcat(buf, "Context Path");
+    else if (match & MATCH_TYPE_SUFFIX)
+        strcat(buf, "Suffix");
+    else if (match & MATCH_TYPE_GENERAL_SUFFIX)
+        return "General Suffix";
+ */
+    else
+        strcat(buf, "Unknown");
+    return buf;
+}
+
 /*
  * Given context uri, count the number of path tokens.
  *
@@ -78,7 +130,10 @@ static int worker_compare(const void *elem1, const void *elem2)
     /* given the same number of URI tokens, use character
      * length as a tie breaker
      */
-    return ((int)e2->context_len - (int)e1->context_len);
+    if(e2->context_len != e1->context_len)
+        return ((int)e2->context_len - (int)e1->context_len);
+
+    return ((int)e2->source_type - (int)e1->source_type);
 }
 
 static void worker_qsort(jk_uri_worker_map_t *uw_map)
@@ -133,7 +188,7 @@ int uri_worker_map_alloc(jk_uri_worker_map_t **uw_map,
         JK_INIT_CS(&((*uw_map)->cs), rc);
         if (rc == JK_FALSE) {
             jk_log(l, JK_LOG_ERROR,
-                   "creating thread lock errno=%d",
+                   "creating thread lock (errno=%d)",
                    errno);
             JK_TRACE_EXIT(l);
             return JK_FALSE;
@@ -217,13 +272,45 @@ static int uri_worker_map_realloc(jk_uri_worker_map_t *uw_map)
 }
 
 
+/*
+ * Delete all entries of a given source type
+ */
+
+int uri_worker_map_clear(jk_uri_worker_map_t *uw_map,
+                         unsigned int source_type, jk_logger_t *l)
+{
+    uri_worker_record_t *uwr = NULL;
+    unsigned int i;
+    unsigned int j;
+
+    JK_TRACE_ENTER(l);
+
+    for (i = 0; i < uw_map->size; i++) {
+        uwr = uw_map->maps[i];
+        if (uwr->source_type == source_type) {
+            jk_log(l, JK_LOG_DEBUG,
+                   "deleting map rule '%s=%s' source '%s'",
+                   uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
+            for (j = i; j < uw_map->size-1; j++)
+                uw_map->maps[j] = uw_map->maps[j+1];
+            uw_map->size--;
+            i--;
+        }
+    }
+
+    JK_TRACE_EXIT(l);
+    return JK_TRUE;
+}
+
+
+
 int uri_worker_map_add(jk_uri_worker_map_t *uw_map,
-                       const char *puri, const char *worker, jk_logger_t *l)
+                       const char *puri, const char *worker,
+                       unsigned int source_type, jk_logger_t *l)
 {
     uri_worker_record_t *uwr = NULL;
     char *uri;
     unsigned int match_type = 0;
-    unsigned int i;
 
     JK_TRACE_ENTER(l);
 
@@ -240,32 +327,6 @@ int uri_worker_map_add(jk_uri_worker_map_t *uw_map,
         puri++;
     }
 
-    /* Find if duplicate entry */
-    for (i = 0; i < uw_map->size; i++) {
-        uwr = uw_map->maps[i];
-        if (strcmp(uwr->uri, puri) == 0) {
-            /* Update disabled flag */
-            if (match_type & MATCH_TYPE_DISABLED)
-                uwr->match_type |= MATCH_TYPE_DISABLED;
-            else
-                uwr->match_type &= ~MATCH_TYPE_DISABLED;
-            if (strcmp(uwr->worker_name, worker) == 0) {
-                jk_log(l, JK_LOG_DEBUG,
-                       "map rule %s=%s already exists",
-                       puri, worker);
-                JK_TRACE_EXIT(l);
-                return JK_TRUE;
-            }
-            else {
-                jk_log(l, JK_LOG_DEBUG,
-                       "changing map rule %s=%s ",
-                       puri, worker);
-                uwr->worker_name = jk_pool_strdup(&uw_map->p, worker);
-                JK_TRACE_EXIT(l);
-                return JK_TRUE;
-            }
-        }
-    }
     if (uri_worker_map_realloc(uw_map) == JK_FALSE) {
         JK_TRACE_EXIT(l);
         return JK_FALSE;
@@ -288,6 +349,11 @@ int uri_worker_map_add(jk_uri_worker_map_t *uw_map,
     }
 
     if (*uri == '/') {
+        uwr->uri = uri;
+        uwr->context = uri;
+        uwr->worker_name = jk_pool_strdup(&uw_map->p, worker);
+        uwr->context_len = strlen(uwr->context);
+        uwr->source_type = source_type;
         if (strchr(uri, '*') ||
             strchr(uri, '?')) {
             /* Something like
@@ -296,21 +362,17 @@ int uri_worker_map_add(jk_uri_worker_map_t *uw_map,
              */
             match_type |= MATCH_TYPE_WILDCHAR_PATH;
             jk_log(l, JK_LOG_DEBUG,
-                    "wildchar rule %s=%s was added",
-                    uri, worker);
+                   "wildchar rule '%s=%s' source '%s' was added",
+                   uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
 
         }
         else {
             /* Something like:  JkMount /login/j_security_check ajp13 */
             match_type |= MATCH_TYPE_EXACT;
             jk_log(l, JK_LOG_DEBUG,
-                   "exact rule %s=%s was added",
-                   uri, worker);
+                   "exact rule '%s=%s' source '%s' was added",
+                   uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
         }
-        uwr->uri = uri;
-        uwr->context = uri;
-        uwr->worker_name = jk_pool_strdup(&uw_map->p, worker);
-        uwr->context_len = strlen(uwr->context);
     }
     else {
         /*
@@ -320,7 +382,7 @@ int uri_worker_map_add(jk_uri_worker_map_t *uw_map,
          * not arrive here when using Apache.
          */
         jk_log(l, JK_LOG_ERROR,
-               "invalid context %s",
+               "invalid context '%s': does not begin with '/'",
                uri);
         JK_TRACE_EXIT(l);
         return JK_FALSE;
@@ -379,7 +441,7 @@ int uri_worker_map_open(jk_uri_worker_map_t *uw_map,
                     s = strchr(r, '|');
                     *(s++) = '\0';
                     /* Add first mapping */
-                    if (!uri_worker_map_add(uw_map, r, w, l)) {
+                    if (!uri_worker_map_add(uw_map, r, w, SOURCE_TYPE_JKMOUNT, l)) {
                         jk_log(l, JK_LOG_ERROR,
                         "invalid mapping rule %s->%s", r, w);
                         rc = JK_FALSE;
@@ -388,14 +450,14 @@ int uri_worker_map_open(jk_uri_worker_map_t *uw_map,
                         *(s - 1) = *s;
                     *(s - 1) = '\0';
                     /* add second mapping */
-                    if (!uri_worker_map_add(uw_map, r, w, l)) {
+                    if (!uri_worker_map_add(uw_map, r, w, SOURCE_TYPE_JKMOUNT, l)) {
                         jk_log(l, JK_LOG_ERROR,
                                "invalid mapping rule %s->%s", r, w);
                         rc = JK_FALSE;
                     }
                     free(r);
                 }
-                else if (!uri_worker_map_add(uw_map, u, w, l)) {
+                else if (!uri_worker_map_add(uw_map, u, w, SOURCE_TYPE_JKMOUNT, l)) {
                     jk_log(l, JK_LOG_ERROR,
                            "invalid mapping rule %s->%s",
                            u, w);
@@ -433,21 +495,21 @@ static int is_nomap_match(jk_uri_worker_map_t *uw_map,
         if (!(uwr->match_type & MATCH_TYPE_NO_MATCH) ||
             (uwr->match_type & MATCH_TYPE_DISABLED))
             continue;
-        /* Check only mathing workers */
+        /* Check only matching workers */
         if (strcmp(uwr->worker_name, worker))
             continue;
         if (uwr->match_type & MATCH_TYPE_WILDCHAR_PATH) {
             /* Map is already sorted by context_len */
             if (wildchar_match(uri, uwr->context,
 #ifdef WIN32
-                               1
+                               0
 #else
                                0
 #endif
                                ) == 0) {
                     jk_log(l, JK_LOG_DEBUG,
-                           "Found a no match %s -> %s",
-                           uwr->worker_name, uwr->context);
+                           "Found a wildchar no match '%s=%s' source '%s'",
+                           uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
                     JK_TRACE_EXIT(l);
                     return JK_TRUE;
              }
@@ -456,8 +518,8 @@ static int is_nomap_match(jk_uri_worker_map_t *uw_map,
             if (strlen(uri) == uwr->context_len) {
                 if (JK_IS_DEBUG_LEVEL(l))
                     jk_log(l, JK_LOG_DEBUG,
-                            "Found an exact no match %s -> %s",
-                            uwr->worker_name, uwr->context);
+                           "Found an exact no match '%s=%s' source '%s'",
+                           uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
                 JK_TRACE_EXIT(l);
                 return JK_TRUE;
             }
@@ -473,12 +535,11 @@ const char *map_uri_to_worker(jk_uri_worker_map_t *uw_map,
                               const char *uri, jk_logger_t *l)
 {
     unsigned int i;
-    char *url_rewrite;
     const char *rv = NULL;
-    const char *url = uri;
-    char  buf[JK_MAX_URI_LEN+1];
+    char  url[JK_MAX_URI_LEN+1];
 
     JK_TRACE_ENTER(l);
+
     if (!uw_map || !uri) {
         JK_LOG_NULL_PARAMS(l);
         JK_TRACE_EXIT(l);
@@ -490,23 +551,43 @@ const char *map_uri_to_worker(jk_uri_worker_map_t *uw_map,
         JK_TRACE_EXIT(l);
         return NULL;
     }
-    url_rewrite = strstr(uri, JK_PATH_SESSION_IDENTIFIER);
-    if (url_rewrite) {
-        size_t len = url_rewrite - uri;
-        if (len > JK_MAX_URI_LEN)
-            len = JK_MAX_URI_LEN;
-        strncpy(buf, uri, len);
-        buf[len] = '\0';
-        url = &buf[0];
-        if (JK_IS_DEBUG_LEVEL(l))
-            jk_log(l, JK_LOG_DEBUG, "Removing Session path '%s' URI '%s'",
-                   url_rewrite, url);
-    }
-    if (uw_map->fname)
+    if (uw_map->fname) {
         uri_worker_map_update(uw_map, l);
+        if (!uw_map->size) {
+            jk_log(l, JK_LOG_INFO,
+                   "No worker maps defined for %s.",
+                   uw_map->fname);
+            JK_TRACE_EXIT(l);
+            return NULL;
+        }
+    }
+    /* Make the copy of the provided uri and strip
+     * everything after the first ';' char.
+     */
+    for (i = 0; i < strlen(uri); i++) {
+        if (i == JK_MAX_URI_LEN) {
+            jk_log(l, JK_LOG_WARNING,
+                   "Uri %s is invalid. Uri must be smaller then %d chars",
+                   uri, JK_MAX_URI_LEN);
+            JK_TRACE_EXIT(l);
+            return NULL;
+        }
+        if (uri[i] == ';')
+            break;
+        else
+            url[i] = uri[i];
+    }
+    url[i] = '\0';
+
+    if (JK_IS_DEBUG_LEVEL(l)) {
+        char *url_rewrite = strstr(uri, JK_PATH_SESSION_IDENTIFIER);
+        if (url_rewrite)
+            jk_log(l, JK_LOG_DEBUG, "separating session identifier '%s' from url '%s'",
+                   url_rewrite, uri);
+    }
     if (JK_IS_DEBUG_LEVEL(l))
         jk_log(l, JK_LOG_DEBUG, "Attempting to map URI '%s' from %d maps",
-               uri, uw_map->size);
+               url, uw_map->size);
 
     for (i = 0; i < uw_map->size; i++) {
         uri_worker_record_t *uwr = uw_map->maps[i];
@@ -517,14 +598,15 @@ const char *map_uri_to_worker(jk_uri_worker_map_t *uw_map,
             continue;
 
         if (JK_IS_DEBUG_LEVEL(l))
-            jk_log(l, JK_LOG_DEBUG, "Attempting to map context URI '%s'", uwr->uri);
+            jk_log(l, JK_LOG_DEBUG, "Attempting to map context URI '%s=%s' source '%s'",
+                   uwr->context, uwr->worker_name, uri_worker_map_get_source(uwr, l));
 
         if (uwr->match_type & MATCH_TYPE_WILDCHAR_PATH) {
             const char *wname;
             /* Map is already sorted by context_len */
             if (wildchar_match(url, uwr->context,
 #ifdef WIN32
-                               1
+                               0
 #else
                                0
 #endif
@@ -532,8 +614,8 @@ const char *map_uri_to_worker(jk_uri_worker_map_t *uw_map,
                     wname = uwr->worker_name;
                     if (JK_IS_DEBUG_LEVEL(l))
                         jk_log(l, JK_LOG_DEBUG,
-                               "Found a wildchar match %s -> %s",
-                               uwr->worker_name, uwr->context);
+                               "Found a wildchar match '%s=%s'",
+                               uwr->context, uwr->worker_name);
                     JK_TRACE_EXIT(l);
                     rv = wname;
                     goto cleanup;
@@ -543,8 +625,8 @@ const char *map_uri_to_worker(jk_uri_worker_map_t *uw_map,
             if (strlen(url) == uwr->context_len) {
                 if (JK_IS_DEBUG_LEVEL(l))
                     jk_log(l, JK_LOG_DEBUG,
-                           "Found an exact match %s -> %s",
-                           uwr->worker_name, uwr->context);
+                           "Found an exact match '%s=%s'",
+                           uwr->context, uwr->worker_name);
                 JK_TRACE_EXIT(l);
                 rv = uwr->worker_name;
                 goto cleanup;
@@ -575,8 +657,13 @@ int uri_worker_map_load(jk_uri_worker_map_t *uw_map,
 
     jk_map_alloc(&map);
     if (jk_map_read_properties(map, uw_map->fname,
-                               &uw_map->modified)) {
+                               &uw_map->modified, 0, l)) {
         int i;
+        if (JK_IS_DEBUG_LEVEL(l))
+            jk_log(l, JK_LOG_DEBUG,
+                   "Loading urimaps from %s with reload check interval %d seconds",
+                   uw_map->fname, uw_map->reload);
+        uri_worker_map_clear(uw_map, SOURCE_TYPE_URIMAP, l);
         for (i = 0; i < jk_map_size(map); i++) {
             const char *u = jk_map_name_at(map, i);
             const char *w = jk_map_value_at(map, i);
@@ -592,7 +679,7 @@ int uri_worker_map_load(jk_uri_worker_map_t *uw_map,
                 s = strchr(r, '|');
                 *(s++) = '\0';
                 /* Add first mapping */
-                if (!uri_worker_map_add(uw_map, r, w, l)) {
+                if (!uri_worker_map_add(uw_map, r, w, SOURCE_TYPE_URIMAP, l)) {
                     jk_log(l, JK_LOG_ERROR,
                     "invalid mapping rule %s->%s", r, w);
                 }
@@ -600,13 +687,13 @@ int uri_worker_map_load(jk_uri_worker_map_t *uw_map,
                    *(s - 1) = *s;
                 *(s - 1) = '\0';
                 /* add second mapping */
-                if (!uri_worker_map_add(uw_map, r, w, l)) {
+                if (!uri_worker_map_add(uw_map, r, w, SOURCE_TYPE_URIMAP, l)) {
                     jk_log(l, JK_LOG_ERROR,
                     "invalid mapping rule %s->%s", r, w);
                 }
                 free(r);
             }
-            else if (!uri_worker_map_add(uw_map, u, w, l)) {
+            else if (!uri_worker_map_add(uw_map, u, w, SOURCE_TYPE_URIMAP, l)) {
                 jk_log(l, JK_LOG_ERROR,
                        "invalid mapping rule %s->%s",
                        u, w);
@@ -625,17 +712,30 @@ int uri_worker_map_update(jk_uri_worker_map_t *uw_map,
     int rc = JK_TRUE;
     time_t now = time(NULL);
 
-    if ((now - uw_map->checked) > JK_URIMAP_RELOAD) {
+    if (uw_map->reload > 0 && difftime(now, uw_map->checked) > uw_map->reload) {
         struct stat statbuf;
         uw_map->checked = now;
-        if ((rc = stat(uw_map->fname, &statbuf)) == -1)
+        if ((rc = stat(uw_map->fname, &statbuf)) == -1) {
+            jk_log(l, JK_LOG_ERROR,
+                   "Unable to stat the %s (errno=%d)",
+                   uw_map->fname, errno);
             return JK_FALSE;
-        if (statbuf.st_mtime == uw_map->modified)
+        }
+        if (statbuf.st_mtime == uw_map->modified) {
+            if (JK_IS_DEBUG_LEVEL(l))
+                jk_log(l, JK_LOG_DEBUG,
+                       "File %s  is not modified",
+                       uw_map->fname);
             return JK_TRUE;
+        }
         JK_ENTER_CS(&(uw_map->cs), rc);
         /* Check if some other thread updated status */
         if (statbuf.st_mtime == uw_map->modified) {
             JK_LEAVE_CS(&(uw_map->cs), rc);
+            if (JK_IS_DEBUG_LEVEL(l))
+                jk_log(l, JK_LOG_DEBUG,
+                       "File %s  is not modified",
+                       uw_map->fname);
             return JK_TRUE;
         }
         rc = uri_worker_map_load(uw_map, l);
@@ -643,6 +743,5 @@ int uri_worker_map_update(jk_uri_worker_map_t *uw_map,
         jk_log(l, JK_LOG_INFO,
                "Reloaded urimaps from %s", uw_map->fname);
     }
-    return rc;
+    return JK_TRUE;
 }
-
