@@ -1,9 +1,10 @@
 /*
- *  Copyright 1999-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -19,7 +20,7 @@
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Henri Gomez <hgomez@apache.org>                            *
  * Author:      Rainer Jung <rjung@apache.org>                             *
- * Version:     $Revision: 411541 $                                          *
+ * Version:     $Revision: 527143 $                                          *
  ***************************************************************************/
 
 
@@ -34,7 +35,6 @@
 #define STDERR_OF_WORKER            ("stderr")
 #define STDOUT_OF_WORKER            ("stdout")
 #define SECRET_OF_WORKER            ("secret")
-#define CONF_OF_WORKER              ("conf")
 #define MX_OF_WORKER                ("mx")
 #define MS_OF_WORKER                ("ms")
 #define CP_OF_WORKER                ("class_path")
@@ -61,33 +61,49 @@
 #define RECYCLE_TIMEOUT_DEPRECATED  ("recycle_timeout")
 #define LOAD_FACTOR_OF_WORKER       ("lbfactor")
 #define DISTANCE_OF_WORKER          ("distance")
-/* deprecated directive. Use balance_workers instead */
-#define BALANCED_WORKERS            ("balanced_workers")
+#define BALANCED_WORKERS_DEPRECATED ("balanced_workers")
 #define BALANCE_WORKERS             ("balance_workers")
 #define STICKY_SESSION              ("sticky_session")
 #define STICKY_SESSION_FORCE        ("sticky_session_force")
-#define JVM_ROUTE_OF_WORKER         ("jvm_route")
+#define LOCAL_WORKER_DEPRECATED     ("local_worker")
+#define LOCAL_WORKER_ONLY_DEPRECATED ("local_worker_only")
+#define JVM_ROUTE_OF_WORKER_DEPRECATED ("jvm_route")
+#define ROUTE_OF_WORKER             ("route")
 #define DOMAIN_OF_WORKER            ("domain")
 #define REDIRECT_OF_WORKER          ("redirect")
 #define MOUNT_OF_WORKER             ("mount")
 #define METHOD_OF_WORKER            ("method")
 #define LOCK_OF_WORKER              ("lock")
-#define IS_WORKER_DISABLED          ("disabled")
-#define IS_WORKER_STOPPED           ("stopped")
+#define IS_WORKER_DISABLED_DEPRECATED ("disabled")
+#define IS_WORKER_STOPPED_DEPRECATED  ("stopped")
+#define ACTIVATION_OF_WORKER        ("activation")
 #define WORKER_RECOVER_TIME         ("recover_time")
+#define WORKER_MAX_PACKET_SIZE      ("max_packet_size")
+#define STYLE_SHEET_OF_WORKER       ("css")
+#define NAMESPACE_OF_WORKER         ("ns")
+#define XML_NAMESPACE_OF_WORKER     ("xmlns")
+#define XML_DOCTYPE_OF_WORKER       ("doctype")
+#define PROP_PREFIX_OF_WORKER       ("prefix")
 
+#define READ_ONLY_OF_WORKER         ("read_only")
+#define USER_OF_WORKER              ("user")
+#define USER_CASE_OF_WORKER         ("user_case_insensitive")
+#define GOOD_RATING_OF_WORKER       ("good")
+#define BAD_RATING_OF_WORKER        ("bad")
 
 #define DEFAULT_WORKER_TYPE         JK_AJP13_WORKER_NAME
 #define SECRET_KEY_OF_WORKER        ("secretkey")
 #define RETRIES_OF_WORKER           ("retries")
+#define STATUS_FAIL_OF_WORKER       ("fail_on_status")
 
 #define DEFAULT_WORKER              JK_AJP13_WORKER_NAME
 #define WORKER_LIST_PROPERTY_NAME     ("worker.list")
+#define LIST_PROPERTY_NAME            ("list")
 #define WORKER_MAINTAIN_PROPERTY_NAME ("worker.maintain")
+#define MAINTAIN_PROPERTY_NAME        ("maintain")
 #define DEFAULT_MAINTAIN_TIME       (60)
 #define DEFAULT_LB_FACTOR           (1)
 #define DEFAULT_DISTANCE            (0)
-#define LOG_FORMAT                  ("log_format")
 
 #define TOMCAT32_BRIDGE_NAME        ("tomcat32")
 #define TOMCAT33_BRIDGE_NAME        ("tomcat33")
@@ -120,6 +136,153 @@
     extern long _ftol2(double dblSource) { return _ftol(dblSource); }
 #endif
 
+static const char *list_properties[] = {
+    BALANCE_WORKERS,
+    MOUNT_OF_WORKER,
+    USER_OF_WORKER,
+    GOOD_RATING_OF_WORKER,
+    BAD_RATING_OF_WORKER,
+    STATUS_FAIL_OF_WORKER,
+    "list",
+    NULL
+};
+
+static const char *unique_properties[] = {
+    SECRET_OF_WORKER,
+    HOST_OF_WORKER,
+    PORT_OF_WORKER,
+    TYPE_OF_WORKER,
+    CACHE_OF_WORKER_DEPRECATED,
+    CACHE_OF_WORKER,
+    CACHE_OF_WORKER_MIN,
+    CACHE_TIMEOUT_DEPRECATED,
+    CACHE_TIMEOUT_OF_WORKER,
+    RECOVERY_OPTS_OF_WORKER,
+    CONNECT_TIMEOUT_OF_WORKER,
+    PREPOST_TIMEOUT_OF_WORKER,
+    REPLY_TIMEOUT_OF_WORKER,
+    SOCKET_TIMEOUT_OF_WORKER,
+    SOCKET_BUFFER_OF_WORKER,
+    SOCKET_KEEPALIVE_OF_WORKER,
+    RECYCLE_TIMEOUT_DEPRECATED,
+    LOAD_FACTOR_OF_WORKER,
+    STICKY_SESSION,
+    STICKY_SESSION_FORCE,
+    LOCAL_WORKER_DEPRECATED,
+    LOCAL_WORKER_ONLY_DEPRECATED,
+    JVM_ROUTE_OF_WORKER_DEPRECATED,
+    ROUTE_OF_WORKER,
+    DOMAIN_OF_WORKER,
+    REDIRECT_OF_WORKER,
+    METHOD_OF_WORKER,
+    LOCK_OF_WORKER,
+    IS_WORKER_DISABLED_DEPRECATED,
+    IS_WORKER_STOPPED_DEPRECATED,
+    ACTIVATION_OF_WORKER,
+    WORKER_RECOVER_TIME,
+    WORKER_MAX_PACKET_SIZE,
+    STYLE_SHEET_OF_WORKER,
+    READ_ONLY_OF_WORKER,
+    RETRIES_OF_WORKER,
+    WORKER_MAINTAIN_PROPERTY_NAME,
+    NAMESPACE_OF_WORKER,
+    XML_NAMESPACE_OF_WORKER,
+    XML_DOCTYPE_OF_WORKER,
+    PROP_PREFIX_OF_WORKER,
+    USER_CASE_OF_WORKER,
+    NULL
+};
+
+static const char *deprecated_properties[] = {
+    SYSPROPS_OF_WORKER,
+    STDERR_OF_WORKER,
+    STDOUT_OF_WORKER,
+    MX_OF_WORKER,
+    MS_OF_WORKER,
+    CP_OF_WORKER,
+    BRIDGE_OF_WORKER,
+    JVM_OF_WORKER,
+    LIBPATH_OF_WORKER,
+    CMD_LINE_OF_WORKER,
+    NATIVE_LIB_OF_WORKER,
+    CACHE_OF_WORKER_DEPRECATED,
+    CACHE_TIMEOUT_DEPRECATED,
+    RECYCLE_TIMEOUT_DEPRECATED,
+    BALANCED_WORKERS_DEPRECATED,
+    JVM_ROUTE_OF_WORKER_DEPRECATED,
+    LOCAL_WORKER_DEPRECATED,
+    LOCAL_WORKER_ONLY_DEPRECATED,
+    IS_WORKER_DISABLED_DEPRECATED,
+    IS_WORKER_STOPPED_DEPRECATED,
+    NULL
+};
+
+static const char *supported_properties[] = {
+    SYSPROPS_OF_WORKER,
+    STDERR_OF_WORKER,
+    STDOUT_OF_WORKER,
+    SECRET_OF_WORKER,
+    MX_OF_WORKER,
+    MS_OF_WORKER,
+    CP_OF_WORKER,
+    BRIDGE_OF_WORKER,
+    JVM_OF_WORKER,
+    LIBPATH_OF_WORKER,
+    CMD_LINE_OF_WORKER,
+    NATIVE_LIB_OF_WORKER,
+    HOST_OF_WORKER,
+    PORT_OF_WORKER,
+    TYPE_OF_WORKER,
+    CACHE_OF_WORKER_DEPRECATED,
+    CACHE_OF_WORKER,
+    CACHE_OF_WORKER_MIN,
+    CACHE_TIMEOUT_DEPRECATED,
+    CACHE_TIMEOUT_OF_WORKER,
+    RECOVERY_OPTS_OF_WORKER,
+    CONNECT_TIMEOUT_OF_WORKER,
+    PREPOST_TIMEOUT_OF_WORKER,
+    REPLY_TIMEOUT_OF_WORKER,
+    SOCKET_TIMEOUT_OF_WORKER,
+    SOCKET_BUFFER_OF_WORKER,
+    SOCKET_KEEPALIVE_OF_WORKER,
+    RECYCLE_TIMEOUT_DEPRECATED,
+    LOAD_FACTOR_OF_WORKER,
+    DISTANCE_OF_WORKER,
+    BALANCED_WORKERS_DEPRECATED,
+    BALANCE_WORKERS,
+    STICKY_SESSION,
+    STICKY_SESSION_FORCE,
+    LOCAL_WORKER_DEPRECATED,
+    LOCAL_WORKER_ONLY_DEPRECATED,
+    JVM_ROUTE_OF_WORKER_DEPRECATED,
+    ROUTE_OF_WORKER,
+    DOMAIN_OF_WORKER,
+    REDIRECT_OF_WORKER,
+    MOUNT_OF_WORKER,
+    METHOD_OF_WORKER,
+    LOCK_OF_WORKER,
+    IS_WORKER_DISABLED_DEPRECATED,
+    IS_WORKER_STOPPED_DEPRECATED,
+    ACTIVATION_OF_WORKER,
+    WORKER_RECOVER_TIME,
+    WORKER_MAX_PACKET_SIZE,
+    STYLE_SHEET_OF_WORKER,
+    NAMESPACE_OF_WORKER,
+    XML_NAMESPACE_OF_WORKER,
+    XML_DOCTYPE_OF_WORKER,
+    PROP_PREFIX_OF_WORKER,
+    READ_ONLY_OF_WORKER,
+    USER_OF_WORKER,
+    USER_CASE_OF_WORKER,
+    GOOD_RATING_OF_WORKER,
+    BAD_RATING_OF_WORKER,
+    SECRET_KEY_OF_WORKER,
+    RETRIES_OF_WORKER,
+    STATUS_FAIL_OF_WORKER,
+    LIST_PROPERTY_NAME,
+    MAINTAIN_PROPERTY_NAME
+};
+
 /* All entries need to have fixed length 8 chars! */
 static const char *jk_level_verbs[] = {
     "[" JK_LOG_TRACE_VERB "] ",
@@ -131,15 +294,62 @@ static const char *jk_level_verbs[] = {
     NULL
 };
 
-const char *jk_log_fmt = JK_TIME_FORMAT;
+const char *jk_get_bool(int v)
+{
+    if (v == 0)
+        return "False";
+    else
+        return "True";
+}
 
-static size_t set_time_str(char *str, int len)
+int jk_get_bool_code(const char *v, int def)
+{
+    if (!v) {
+        return def;
+    }
+    else if (!strcasecmp(v, "off") ||
+             *v == 'F' || *v == 'f' ||
+             *v == 'N' || *v == 'n' ||
+             *v == '0') {
+        return 0;
+    }
+    else if (!strcasecmp(v, "on") ||
+             *v == 'T' || *v == 't' ||
+             *v == 'Y' || *v == 'y' ||
+             *v == '1') {
+        return 1;
+    }
+    return def;
+}
+
+/* Sleep for 100ms */
+void jk_sleep(int ms)
+{
+#ifdef OS2
+    DosSleep(ms);
+#elif defined(BEOS)
+    snooze(ms * 1000);
+#elif defined(NETWARE)
+    delay(ms);
+#elif defined(WIN32)
+    Sleep(ms);
+#else
+    struct timeval tv;
+    tv.tv_usec = ms * 1000;
+    tv.tv_sec = 0;
+    select(0, NULL, NULL, NULL, &tv);
+#endif
+}
+
+static int set_time_str(char *str, int len, const char *jk_log_fmt)
 {
     time_t t = time(NULL);
     struct tm *tms;
 
     tms = localtime(&t);
-    return strftime(str, len, jk_log_fmt, tms);
+    if (jk_log_fmt)
+        return (int)strftime(str, len, jk_log_fmt, tms);
+    return (int)strftime(str, len, JK_TIME_FORMAT, tms);
 }
 
 /* Write at most n characters to the buffer in str, return the
@@ -187,17 +397,15 @@ static int JK_METHOD log_to_file(jk_logger_t *l, int level, const char *what)
     if (l &&
         (l->level <= level || level == JK_LOG_REQUEST_LEVEL) &&
         l->logger_private && what) {
-        size_t sz = strlen(what);
-        if (sz) {
-            file_logger_t *p = l->logger_private;
-            fwrite(what, 1, sz, p->logfile);
+        jk_file_logger_t *p = l->logger_private;
+        if (p->logfile) {
+            fputs(what, p->logfile);
+            fputc('\n', p->logfile);
             /* [V] Flush the dam' thing! */
             fflush(p->logfile);
         }
-
         return JK_TRUE;
     }
-
     return JK_FALSE;
 }
 
@@ -227,17 +435,19 @@ int jk_parse_log_level(const char *level)
         return JK_LOG_EMERG_LEVEL;
     }
 
-    return JK_LOG_INFO_LEVEL;
+    return JK_LOG_DEF_LEVEL;
 }
 
 int jk_open_file_logger(jk_logger_t **l, const char *file, int level)
 {
     if (l && file) {
+
         jk_logger_t *rc = (jk_logger_t *)malloc(sizeof(jk_logger_t));
-        file_logger_t *p = (file_logger_t *) malloc(sizeof(file_logger_t));
+        jk_file_logger_t *p = (jk_file_logger_t *) malloc(sizeof(jk_file_logger_t));
         if (rc && p) {
             rc->log = log_to_file;
             rc->level = level;
+            rc->log_fmt = NULL;
             rc->logger_private = p;
 #ifdef AS400
             p->logfile = fopen(file, "a+, o_ccsid=0");
@@ -264,7 +474,7 @@ int jk_open_file_logger(jk_logger_t **l, const char *file, int level)
 int jk_close_file_logger(jk_logger_t **l)
 {
     if (l && *l) {
-        file_logger_t *p = (*l)->logger_private;
+        jk_file_logger_t *p = (*l)->logger_private;
         if (p) {
             fflush(p->logfile);
             fclose(p->logfile);
@@ -283,8 +493,8 @@ int jk_log(jk_logger_t *l,
            const char *fmt, ...)
 {
     int rc = 0;
-    /* Need to reserve space for newline and terminating zero byte. */
-    static int usable_size = HUGE_BUFFER_SIZE-2;
+    /* Need to reserve space for terminating zero byte. */
+    static int usable_size = HUGE_BUFFER_SIZE - 1;
     if (!l || !file || !fmt) {
         return -1;
     }
@@ -313,7 +523,7 @@ int jk_log(jk_logger_t *l,
         if (NULL == buf)
             return -1;
 #endif
-        used = set_time_str(buf, usable_size);
+        used = set_time_str(buf, usable_size, l->log_fmt);
 
         if (line) {
             /* Log [pid:threadid] for all levels except REQUEST. */
@@ -335,7 +545,7 @@ int jk_log(jk_logger_t *l,
             used += 8;
 
             if (funcname) {
-                rc = strlen(funcname) + 2;
+                rc = (int)strlen(funcname) + 2;
                 if (usable_size - used >= rc) {
                     strcat(buf, funcname);
                     strcat(buf, "::");
@@ -361,14 +571,13 @@ int jk_log(jk_logger_t *l,
 #else
         rc = vsnprintf(buf + used, usable_size - used, fmt, args);
 #endif
+        va_end(args);
         if ( rc <= usable_size - used ) {
             used += rc;
         } else {
             used = usable_size;
         }
-        buf[used] = '\n';
-        buf[used+1] = 0;
-        va_end(args);
+        buf[used] = 0;
         l->log(l, level, buf);
 #ifdef NETWARE
         free(buf);
@@ -389,13 +598,20 @@ const char *jk_get_worker_type(jk_map_t *m, const char *wname)
     return jk_map_get_string(m, buf, DEFAULT_WORKER_TYPE);
 }
 
-const char *jk_get_worker_jvm_route(jk_map_t *m, const char *wname, const char *def)
+const char *jk_get_worker_route(jk_map_t *m, const char *wname, const char *def)
 {
     char buf[1024];
+    const char *v;
     if (!m || !wname) {
         return NULL;
     }
-    MAKE_WORKER_PARAM(JVM_ROUTE_OF_WORKER);
+    MAKE_WORKER_PARAM(ROUTE_OF_WORKER);
+    v = jk_map_get_string(m, buf, NULL);
+    if (v) {
+        return v;
+    }
+    /* Try old jvm_route directive */
+    MAKE_WORKER_PARAM(JVM_ROUTE_OF_WORKER_DEPRECATED);
     return jk_map_get_string(m, buf, def);
 }
 
@@ -697,19 +913,19 @@ const char *jk_get_worker_secret_key(jk_map_t *m, const char *wname)
     return jk_map_get_string(m, buf, NULL);
 }
 
-int jk_get_worker_list(jk_map_t *m, char ***list, unsigned *num_of_wokers)
+int jk_get_worker_list(jk_map_t *m, char ***list, unsigned *num_of_workers)
 {
-    if (m && list && num_of_wokers) {
+    if (m && list && num_of_workers) {
         char **ar = jk_map_get_string_list(m,
                                         WORKER_LIST_PROPERTY_NAME,
-                                        num_of_wokers,
+                                        num_of_workers,
                                         DEFAULT_WORKER);
         if (ar) {
             *list = ar;
             return JK_TRUE;
         }
         *list = NULL;
-        *num_of_wokers = 0;
+        *num_of_workers = 0;
     }
 
     return JK_FALSE;
@@ -721,7 +937,7 @@ int jk_get_is_worker_disabled(jk_map_t *m, const char *wname)
     char buf[1024];
     if (m && wname) {
         int value;
-        MAKE_WORKER_PARAM(IS_WORKER_DISABLED);
+        MAKE_WORKER_PARAM(IS_WORKER_DISABLED_DEPRECATED);
         value = jk_map_get_bool(m, buf, 0);
         if (!value)
             rc = JK_FALSE;
@@ -735,7 +951,7 @@ int jk_get_is_worker_stopped(jk_map_t *m, const char *wname)
     char buf[1024];
     if (m && wname) {
         int value;
-        MAKE_WORKER_PARAM(IS_WORKER_STOPPED);
+        MAKE_WORKER_PARAM(IS_WORKER_STOPPED_DEPRECATED);
         value = jk_map_get_bool(m, buf, 0);
         if (!value)
             rc = JK_FALSE;
@@ -743,9 +959,25 @@ int jk_get_is_worker_stopped(jk_map_t *m, const char *wname)
     return rc;
 }
 
-void jk_set_log_format(const char *logformat)
+int jk_get_worker_activation(jk_map_t *m, const char *wname)
 {
-    jk_log_fmt = (logformat) ? logformat : JK_TIME_FORMAT;
+    char buf[1024];
+    const char *v;
+    if (!m || !wname) {
+        return JK_LB_ACTIVATION_ACTIVE;
+    }
+
+    MAKE_WORKER_PARAM(ACTIVATION_OF_WORKER);
+    v = jk_map_get_string(m, buf, NULL);
+    if (v) {
+        return jk_lb_get_activation_code(v);
+    }
+    else if (jk_get_is_worker_stopped(m, wname))
+        return JK_LB_ACTIVATION_STOPPED;
+    else if (jk_get_is_worker_disabled(m, wname))
+        return JK_LB_ACTIVATION_DISABLED;
+    else
+        return JK_LB_ACTIVATION_DEF;
 }
 
 int jk_get_lb_factor(jk_map_t *m, const char *wname)
@@ -807,21 +1039,12 @@ int jk_get_lb_method(jk_map_t *m, const char *wname)
     char buf[1024];
     const char *v;
     if (!m || !wname) {
-        return JK_LB_BYREQUESTS;
+        return JK_LB_METHOD_DEF;
     }
 
     MAKE_WORKER_PARAM(METHOD_OF_WORKER);
     v = jk_map_get_string(m, buf, NULL);
-    if (!v)
-        return JK_LB_BYREQUESTS;
-    else if  (*v == 't' || *v == 'T' || *v == '1')
-        return JK_LB_BYTRAFFIC;
-    else if  (*v == 'r' || *v == 'R' || *v == '0')
-        return JK_LB_BYREQUESTS;
-    else if  (*v == 'b' || *v == 'B' || *v == '2')
-        return JK_LB_BYBUSYNESS;
-    else
-        return JK_LB_BYREQUESTS;
+    return jk_lb_get_method_code(v);
 }
 
 int jk_get_lb_lock(jk_map_t *m, const char *wname)
@@ -829,45 +1052,233 @@ int jk_get_lb_lock(jk_map_t *m, const char *wname)
     char buf[1024];
     const char *v;
     if (!m || !wname) {
-        return JK_LB_LOCK_DEFAULT;
+        return JK_LB_LOCK_DEF;
     }
 
     MAKE_WORKER_PARAM(LOCK_OF_WORKER);
     v = jk_map_get_string(m, buf, NULL);
-    if (!v)
-        return JK_LB_LOCK_DEFAULT;
-    else if  (*v == 'o' || *v == 'O' || *v == '0')
-        return JK_LB_LOCK_DEFAULT;
-    else if  (*v == 'p' || *v == 'P' || *v == '1')
-        return JK_LB_LOCK_PESSIMISTIC;
-    else
-        return JK_LB_LOCK_DEFAULT;
+    return jk_lb_get_lock_code(v);
 }
 
-int jk_get_lb_worker_list(jk_map_t *m,
-                          const char *wname,
-                          char ***list, unsigned int *num_of_wokers)
+int jk_get_max_packet_size(jk_map_t *m, const char *wname)
+{
+    char buf[1024];
+    int sz;
+
+    if (!m || !wname) {
+        return DEF_BUFFER_SZ;
+    }
+
+    MAKE_WORKER_PARAM(WORKER_MAX_PACKET_SIZE);
+    sz = jk_map_get_int(m, buf, DEF_BUFFER_SZ);
+    sz = JK_ALIGN(sz, 1024);
+    if (sz < DEF_BUFFER_SZ)
+        sz = DEF_BUFFER_SZ;
+    else if (sz > 64*1024)
+        sz = 64*1024;
+
+    return sz;
+}
+
+int jk_get_worker_fail_on_status(jk_map_t *m, const char *wname,
+                                 int *list, unsigned int list_size)
+{
+    char buf[1024];
+    if (!m || !wname || !list) {
+        return 0;
+    }
+    MAKE_WORKER_PARAM(STATUS_FAIL_OF_WORKER);
+    if (list_size) {
+        return jk_map_get_int_list(m, buf,
+                                   list, list_size,
+                                   NULL);
+    }
+
+    return 0;
+}
+
+int jk_get_worker_user_case_insensitive(jk_map_t *m, const char *wname)
+{
+    int rc = JK_FALSE;
+    char buf[1024];
+    if (m && wname) {
+        int value;
+        MAKE_WORKER_PARAM(USER_CASE_OF_WORKER);
+        value = jk_map_get_bool(m, buf, 0);
+        if (value)
+            rc = JK_TRUE;
+    }
+    return rc;
+
+}
+
+const char *jk_get_worker_style_sheet(jk_map_t *m, const char *wname, const char *def)
 {
     char buf[1024];
 
-    if (m && list && num_of_wokers && wname) {
+    if (!m || !wname) {
+        return NULL;
+    }
+
+    MAKE_WORKER_PARAM(STYLE_SHEET_OF_WORKER);
+
+    return jk_map_get_string(m, buf, def);
+}
+
+const char *jk_get_worker_name_space(jk_map_t *m, const char *wname, const char *def)
+{
+    const char *rc;
+    char buf[1024];
+    if (!m || !wname) {
+        return NULL;
+    }
+    MAKE_WORKER_PARAM(NAMESPACE_OF_WORKER);
+    rc = jk_map_get_string(m, buf, def);
+    if (*rc == '-')
+        return "";
+    else
+        return rc;
+}
+
+const char *jk_get_worker_xmlns(jk_map_t *m, const char *wname, const char *def)
+{
+    const char *rc;
+    char buf[1024];
+    if (!m || !wname) {
+        return NULL;
+    }
+    MAKE_WORKER_PARAM(XML_NAMESPACE_OF_WORKER);
+    rc = jk_map_get_string(m, buf, def);
+    if (*rc == '-')
+        return "";
+    else
+        return rc;
+}
+
+const char *jk_get_worker_xml_doctype(jk_map_t *m, const char *wname, const char *def)
+{
+    char buf[1024];
+    if (!m || !wname) {
+        return NULL;
+    }
+    MAKE_WORKER_PARAM(XML_DOCTYPE_OF_WORKER);
+    return jk_map_get_string(m, buf, def);
+}
+
+const char *jk_get_worker_prop_prefix(jk_map_t *m, const char *wname, const char *def)
+{
+    char buf[1024];
+    if (!m || !wname) {
+        return NULL;
+    }
+    MAKE_WORKER_PARAM(PROP_PREFIX_OF_WORKER);
+    return jk_map_get_string(m, buf, def);
+}
+
+int jk_get_is_read_only(jk_map_t *m, const char *wname)
+{
+    int rc = JK_FALSE;
+    char buf[1024];
+    if (m && wname) {
+        int value;
+        MAKE_WORKER_PARAM(READ_ONLY_OF_WORKER);
+        value = jk_map_get_bool(m, buf, 0);
+        if (value)
+            rc = JK_TRUE;
+    }
+    return rc;
+}
+
+int jk_get_worker_user_list(jk_map_t *m,
+                            const char *wname,
+                            char ***list, unsigned int *num)
+{
+    char buf[1024];
+
+    if (m && list && num && wname) {
         char **ar = NULL;
 
-        MAKE_WORKER_PARAM(BALANCE_WORKERS);
-        ar = jk_map_get_string_list(m, buf, num_of_wokers, NULL);
-        if (ar) {
-            *list = ar;
-            return JK_TRUE;
-        }
-        /* Try old balanced_workers directive */
-        MAKE_WORKER_PARAM(BALANCED_WORKERS);
-        ar = jk_map_get_string_list(m, buf, num_of_wokers, NULL);
+        MAKE_WORKER_PARAM(USER_OF_WORKER);
+        ar = jk_map_get_string_list(m, buf, num, NULL);
         if (ar) {
             *list = ar;
             return JK_TRUE;
         }
         *list = NULL;
-        *num_of_wokers = 0;
+        *num = 0;
+    }
+
+    return JK_FALSE;
+}
+
+int jk_get_worker_good_rating(jk_map_t *m,
+                              const char *wname,
+                              char ***list, unsigned int *num)
+{
+    char buf[1024];
+
+    if (m && list && num && wname) {
+        char **ar = NULL;
+
+        MAKE_WORKER_PARAM(GOOD_RATING_OF_WORKER);
+        ar = jk_map_get_string_list(m, buf, num, NULL);
+        if (ar) {
+            *list = ar;
+            return JK_TRUE;
+        }
+        *list = NULL;
+        *num = 0;
+    }
+
+    return JK_FALSE;
+}
+
+int jk_get_worker_bad_rating(jk_map_t *m,
+                             const char *wname,
+                             char ***list, unsigned int *num)
+{
+    char buf[1024];
+
+    if (m && list && num && wname) {
+        char **ar = NULL;
+
+        MAKE_WORKER_PARAM(BAD_RATING_OF_WORKER);
+        ar = jk_map_get_string_list(m, buf, num, NULL);
+        if (ar) {
+            *list = ar;
+            return JK_TRUE;
+        }
+        *list = NULL;
+        *num = 0;
+    }
+
+    return JK_FALSE;
+}
+
+int jk_get_lb_worker_list(jk_map_t *m,
+                          const char *wname,
+                          char ***list, unsigned int *num_of_workers)
+{
+    char buf[1024];
+
+    if (m && list && num_of_workers && wname) {
+        char **ar = NULL;
+
+        MAKE_WORKER_PARAM(BALANCE_WORKERS);
+        ar = jk_map_get_string_list(m, buf, num_of_workers, NULL);
+        if (ar) {
+            *list = ar;
+            return JK_TRUE;
+        }
+        /* Try old balanced_workers directive */
+        MAKE_WORKER_PARAM(BALANCED_WORKERS_DEPRECATED);
+        ar = jk_map_get_string_list(m, buf, num_of_workers, NULL);
+        if (ar) {
+            *list = ar;
+            return JK_TRUE;
+        }
+        *list = NULL;
+        *num_of_workers = 0;
     }
 
     return JK_FALSE;
@@ -1047,14 +1458,20 @@ int jk_file_exists(const char *f)
     return JK_FALSE;
 }
 
-static int jk_is_some_property(const char *prp_name, const char *suffix)
+static int jk_is_some_property(const char *prp_name, const char *suffix, const char *sep)
 {
+    char buf[1024];
+
     if (prp_name && suffix) {
-        size_t prp_name_len = strlen(prp_name);
-        size_t suffix_len = strlen(suffix);
+        size_t prp_name_len;
+        size_t suffix_len;
+        strcpy(buf, sep);
+        strcat(buf, suffix);
+        prp_name_len = strlen(prp_name);
+        suffix_len = strlen(buf);
         if (prp_name_len >= suffix_len) {
             const char *prp_suffix = prp_name + prp_name_len - suffix_len;
-            if (0 == strcmp(suffix, prp_suffix)) {
+            if (0 == strcmp(buf, prp_suffix)) {
                 return JK_TRUE;
             }
         }
@@ -1063,70 +1480,32 @@ static int jk_is_some_property(const char *prp_name, const char *suffix)
     return JK_FALSE;
 }
 
-int jk_is_path_poperty(const char *prp_name)
+int jk_is_path_property(const char *prp_name)
 {
-    return jk_is_some_property(prp_name, "path");
+    return jk_is_some_property(prp_name, "path", "_");
 }
 
-int jk_is_cmd_line_poperty(const char *prp_name)
+int jk_is_cmd_line_property(const char *prp_name)
 {
-    return jk_is_some_property(prp_name, CMD_LINE_OF_WORKER);
+    return jk_is_some_property(prp_name, CMD_LINE_OF_WORKER, ".");
 }
 
-static const char *unique_properties[] = {
-    HOST_OF_WORKER,
-    PORT_OF_WORKER,
-    TYPE_OF_WORKER,
-    CACHE_OF_WORKER,
-    CACHE_OF_WORKER_DEPRECATED,
-    CACHE_OF_WORKER,
-    CACHE_OF_WORKER_MIN,
-    CACHE_TIMEOUT_OF_WORKER,
-    CACHE_TIMEOUT_DEPRECATED,
-    RECOVERY_OPTS_OF_WORKER,
-    CONNECT_TIMEOUT_OF_WORKER,
-    PREPOST_TIMEOUT_OF_WORKER,
-    REPLY_TIMEOUT_OF_WORKER,
-    SOCKET_TIMEOUT_OF_WORKER,
-    SOCKET_BUFFER_OF_WORKER,
-    SOCKET_KEEPALIVE_OF_WORKER,
-    RECYCLE_TIMEOUT_DEPRECATED,
-    LOAD_FACTOR_OF_WORKER,
-    STICKY_SESSION,
-    STICKY_SESSION_FORCE,
-    JVM_ROUTE_OF_WORKER,
-    DOMAIN_OF_WORKER,
-    REDIRECT_OF_WORKER,
-    MOUNT_OF_WORKER,
-    METHOD_OF_WORKER,
-    IS_WORKER_DISABLED,
-    IS_WORKER_STOPPED,
-    WORKER_RECOVER_TIME,
-    SECRET_OF_WORKER,
-    RETRIES_OF_WORKER,
-    NULL
-};
-
-static const char *deprecated_properties[] = {
-    CACHE_OF_WORKER_DEPRECATED,
-    CACHE_TIMEOUT_DEPRECATED,
-    RECYCLE_TIMEOUT_DEPRECATED,
-    MX_OF_WORKER,
-    MS_OF_WORKER,
-    CP_OF_WORKER,
-    BRIDGE_OF_WORKER,
-    JVM_OF_WORKER,
-    LIBPATH_OF_WORKER,
-    CMD_LINE_OF_WORKER,
-    NATIVE_LIB_OF_WORKER,
-    NULL
-};
+int jk_is_list_property(const char *prp_name)
+{
+    const char **props = &list_properties[0];
+    while (*props) {
+        if (jk_is_some_property(prp_name, *props, "."))
+            return JK_TRUE;
+        props++;
+    }
+    return JK_FALSE;
+}
 
 int jk_is_unique_property(const char *prp_name)
 {
     const char **props = &unique_properties[0];
     while (*props) {
-        if (jk_is_some_property(prp_name, *props))
+        if (jk_is_some_property(prp_name, *props, "."))
             return JK_TRUE;
         props++;
     }
@@ -1137,7 +1516,25 @@ int jk_is_deprecated_property(const char *prp_name)
 {
     const char **props = &deprecated_properties[0];
     while (*props) {
-        if (jk_is_some_property(prp_name, *props))
+        if (jk_is_some_property(prp_name, *props, "."))
+            return JK_TRUE;
+        props++;
+    }
+    return JK_FALSE;
+}
+/*
+ * Check that property is a valid one (to prevent user typos).
+ * Only property starting with worker.
+ */
+int jk_is_valid_property(const char *prp_name)
+{
+    const char **props;
+    if (memcmp(prp_name, "worker.", 7))
+        return JK_TRUE;
+
+    props = &supported_properties[0];
+    while (*props) {
+        if (jk_is_some_property(prp_name, *props, "."))
             return JK_TRUE;
         props++;
     }
@@ -1308,8 +1705,9 @@ void jk_init_ws_service(jk_ws_service_t *s)
     s->attributes_names = NULL;
     s->attributes_values = NULL;
     s->num_attributes = 0;
-    s->jvm_route = NULL;
+    s->route = NULL;
     s->retries = JK_RETRIES;
+    s->add_log_items = NULL;
 }
 
 #ifdef _MT_CODE_PTHREAD
