@@ -10,32 +10,80 @@
 # gpg
 # And any one of: w3m, elinks, links
 
-export ANT_HOME=/usr/local/ant
-export JAVA_HOME=/usr/local/jdk1.4.2
+usage() {
+    echo "Usage:: $0 -t VERSION [-b BRANCH | -T]"
+    echo "        -t: version to package"
+    echo "        -b: package from branch BRANCH"
+    echo "        -T: package from trunk"
+}
 
-# You need to change the version numbers
-JK_VERMAJOR="1"
-JK_VERMINOR="2"
-JK_VERFIX="18"
-ASFROOT="http://svn.apache.org/repos/asf"
+while getopts :t:b:T c
+do
+    case $c in
+    t)         tag=$OPTARG;;
+    b)         branch=$OPTARG;;
+    T)         trunk=trunk;;
+    \:)        usage
+               exit 2;;
+    \?)        usage
+               exit 2;;
+    esac
+done
+shift `expr $OPTIND - 1`
+
+SVNROOT="http://svn.apache.org/repos/asf"
+SVNPROJ="tomcat/connectors"
 JK_CVST="tomcat-connectors"
 
-JK_OWNER="asf"
-JK_GROUP="asf"
+JK_OWNER="root"
+JK_GROUP="bin"
 
 COPY_TOP="KEYS LICENSE NOTICE"
-COPY_JK="BUILD.txt native support tools xdocs"
+COPY_JK="BUILD.txt native jkstatus support tools xdocs"
 COPY_CONF="uriworkermap.properties workers.properties workers.properties.minimal"
 
-JK_VER="${JK_VERMAJOR}.${JK_VERMINOR}.${JK_VERFIX}"
-JK_BRANCH="jk${JK_VERMAJOR}.${JK_VERMINOR}.x"
-JK_TAG="JK_${JK_VERMAJOR}_${JK_VERMINOR}_${JK_VERFIX}"
-
-JK_DIST=${JK_CVST}-${JK_VER}-src
-JK_SVN_URL="${ASFROOT}/tomcat/connectors/tags/${JK_BRANCH}/${JK_TAG}"
-
 #################### NO CHANGE BELOW THIS LINE ##############
 
+if [ -z "$tag" ]
+then
+    usage
+    exit 2
+fi
+if [ -n "$trunk" -a -n "$branch" ]
+then
+    usage
+    echo "Only one of the options '-b' and '-T' allowed."
+    exit 2
+fi
+if [ -n "$trunk" ]
+then
+    JK_SVN_URL="${SVNROOT}/${SVNPROJ}/trunk"
+    JK_REV=`svn info ${JK_SVN_URL} | awk '$1 == "Revision:" {print $2}'`
+    if [ -z "$JK_REV" ]
+    then
+       echo "No Revision found at '$JK_SVN_URL'"
+       exit 3
+    fi
+    JK_DIST=${JK_CVST}-${tag}-dev-${JK_REV}-src
+elif [ -n "$branch" ]
+then
+    JK_BRANCH=`echo $branch | sed -e 's#/#__#g'`
+    JK_SVN_URL="${SVNROOT}/${SVNPROJ}/branches/$branch"
+    JK_REV=`svn info ${JK_SVN_URL} | awk '$1 == "Revision:" {print $2}'`
+    if [ -z "$JK_REV" ]
+    then
+       echo "No Revision found at '$JK_SVN_URL'"
+       exit 3
+    fi
+    JK_DIST=${JK_CVST}-${tag}-dev-${JK_BRANCH}-${JK_REV}-src
+else
+    JK_VER=$tag
+    JK_TAG=`echo $tag | sed -e 's#^#JK_#' -e 's#\.#_#g'`
+    JK_BRANCH=`echo $tag | sed -e 's#^#jk#' -e 's#\.[0-9][0-9]*$##' -e 's#$#.x#'`
+    JK_SVN_URL="${SVNROOT}/${SVNPROJ}/tags/${JK_BRANCH}/${JK_TAG}"
+    JK_DIST=${JK_CVST}-${JK_VER}-src
+fi
+
 umask 022
 
 rm -rf ${JK_DIST}
@@ -103,8 +151,8 @@ cd ${JK_DIST}/native
 
 # Check for links, elinks or w3m
 W3MOPTS="-dump -cols 80 -t 4 -S -O iso-8859-1 -T text/html"
-LNKOPTS="-dump"
-ELNKOPTS="--dump --no-numbering --no-home"
+LNKOPTS="-dump -dump-width 80 -dump-charset iso-8859-1 -no-numbering -no-references -no-home"
+ELNKOPTS=$LNKOPTS
 failed=true
 for tool in `echo "w3m elinks links"`
 do
@@ -133,7 +181,8 @@ do
         ;;
     esac
     rm -f CHANGES
-    ${TOOL} ../docs/printer/changelog.html > CHANGES 2>/dev/null
+    echo "Creating the CHANGES file using '$TOOL' ..."
+    ${TOOL} ../docs/miscellaneous/printer/changelog.html > CHANGES 2>/dev/null
     if [ -f CHANGES -a -s CHANGES ]
     then
       failed=false
@@ -148,18 +197,28 @@ then
 fi
 
 # Export text docs
-${TOOL} ../docs/printer/changelog.html >CHANGES
-${TOOL} ../docs/news/printer/20060101.html >NEWS
-${TOOL} ../docs/news/printer/20050101.html >>NEWS
-${TOOL} ../docs/news/printer/20041100.html >>NEWS
+echo "Creating the NEWS file using '$TOOL' ..."
+rm -f NEWS
+touch NEWS
+for news in `ls -r ../xdocs/news/[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9].xml`
+do
+  print=`echo $news | sed -e 's#xdocs/news#docs/news/printer#' -e 's#\.xml#.html#'`
+  echo "Adding $print to NEWS file ..."
+  ${TOOL} $print >>NEWS
+done
+if [ ! -s NEWS ]
+then
+  echo "Can't convert html to text (NEWS)"
+  exit 1
+fi
 
 # Generate configure et. al.
 ./buildconf.sh
 cd ../../
 
 # Pack and sign
-tar cvf ${JK_DIST}.tar --owner="${JK_OWNER}" --group="${JK_GROUP}" ${JK_DIST}
-gzip ${JK_DIST}.tar
+tar cfz ${JK_DIST}.tar.gz --owner="${JK_OWNER}" --group="${JK_GROUP}" ${JK_DIST}
+perl ${JK_DIST}/tools/lineends.pl --cr ${JK_DIST}
 zip -9 -r ${JK_DIST}.zip ${JK_DIST}
 # Create detatched signature
 gpg -ba ${JK_DIST}.tar.gz
